@@ -16,37 +16,77 @@ import argparse
 import json
 import os
 import sys
+
+import jinja2
+from jinja2 import StrictUndefined
+
 import yaml
 
-import tabulate
-
 import directord
+
+from directord import bootstrap
+from directord import client
+from directord import meta
 from directord import mixin
+from directord import server
 from directord import user
 from directord import utils
 
 
-def _args():
+def _args(exec_args=None):
     """Setup client arguments."""
 
     parser = argparse.ArgumentParser(
-        description="Deployment Framework Next.", prog="Directord"
+        description="Deployment Framework Next.",
+        prog="Directord",
+        formatter_class=lambda prog: argparse.HelpFormatter(
+            prog, max_help_position=32, width=128
+        ),
+    )
+    parser.add_argument(
+        "--version",
+        action="version",
+        version="%(prog)s {version}".format(version=directord.__version__),
     )
     parser.add_argument(
         "--config-file",
         help="File path for client configuration. Default: %(default)s",
         metavar="STRING",
-        default=os.getenv(
-            "DIRECTORD_CONFIG_FILE", "/etc/directord/config.yaml"
-        ),
+        default=os.getenv("DIRECTORD_CONFIG_FILE", None),
         type=argparse.FileType(mode="r"),
     )
+    parser.add_argument(
+        "--datastore",
+        help=(
+            "Connect to an external datastore for job and worker tracking. The"
+            " connection string is RFC-1738 compatible >"
+            " driver://username:password@host:port/database. If undefined the"
+            " datastore uses an internal manager object. Driver"
+            " supports ['redis', 'file']. When using Redis, two"
+            " keyspaces will be used and are incremented by 1 using the"
+            " database value (which has a default of 0). If the datastore"
+            " option is set to 'memory', the server will spawn a manager"
+            " thread to facilitate the document store."
+        ),
+        metavar="STRING",
+        default=os.getenv(
+            "DIRECTORD_DATASTORE", "file:///var/cache/directord"
+        ),
+        type=str,
+    )
+    parser.add_argument(
+        "--driver",
+        help="Messaging driver used for workload transport.",
+        default=os.getenv("DIRECTORD_DRIVER", meta.__driver_default__),
+        choices=meta.__driver_options__,
+        type=str,
+    )
     auth_group = parser.add_mutually_exclusive_group()
     auth_group.add_argument(
         "--shared-key",
         help="Shared key used for server client authentication.",
         metavar="STRING",
-        default=os.getenv("DIRECTORD_SHARED_KEY"),
+        default=os.getenv("DIRECTORD_SHARED_KEY", None),
     )
     auth_group.add_argument(
         "--curve-encryption",
@@ -70,17 +110,10 @@ def _args():
         type=int,
     )
     parser.add_argument(
-        "--transfer-port",
-        help="Transfer port to bind. Default: %(default)s",
+        "--backend-port",
+        help="Backend port to bind. Default: %(default)s",
         metavar="INT",
-        default=int(os.getenv("DIRECTORD_TRANSFER_PORT", 5556)),
-        type=int,
-    )
-    parser.add_argument(
-        "--heartbeat-port",
-        help="heartbeat port to bind. Default: %(default)s",
-        metavar="INT",
-        default=int(os.getenv("DIRECTORD_HEARTBEAT_PORT", 5557)),
+        default=int(os.getenv("DIRECTORD_BACKEND_PORT", 5556)),
         type=int,
     )
     parser.add_argument(
@@ -102,9 +135,18 @@ def _args():
         ),
         type=str,
     )
+    parser.add_argument(
+        "--socket-group",
+        help=(
+            "Server file socket group ownership for user interactions."
+            " Default: %(default)s"
+        ),
+        metavar="STRING",
+        default=str(os.getenv("DIRECTORD_SOCKET_GROUP", 0)),
+    )
     parser.add_argument(
         "--cache-path",
-        help=("Client cache path." " Default: %(default)s"),
+        help=("Client cache path. Default: %(default)s"),
         metavar="STRING",
         default=str(os.getenv("DIRECTORD_CACHE_PATH", "/var/cache/directord")),
         type=str,
@@ -117,7 +159,7 @@ def _args():
     )
     parser_orchestrate.add_argument(
         "--restrict",
-        help="Restrict orchestration to a set of Task SHA1(s).",
+        help="Restrict orchestration to a set of Task SHA3_224(s).",
         metavar="STRING",
         nargs="+",
     )
@@ -136,15 +178,40 @@ def _args():
         action="store_true",
     )
     parser_orchestrate.add_argument(
+        "--force-async",
+        help=(
+            "Instruct the orchestration engine to run all orchestrations"
+            " asynchronously."
+        ),
+        action="store_true",
+    )
+    orchestrate_group = parser_orchestrate.add_mutually_exclusive_group(
+        required=False
+    )
+    orchestrate_group.add_argument(
         "--poll",
         help="Block on client return for the completion of executed jobs.",
         action="store_true",
     )
-    parser_orchestrate.add_argument(
+    orchestrate_group.add_argument(
         "--finger-print",
         help="Finger print a set of orchestrations.",
         action="store_true",
     )
+    orchestrate_group.add_argument(
+        "--stream",
+        help="Stream the STDOUT|STDERR for tasks.",
+        action="store_true",
+    )
+    parser_orchestrate.add_argument(
+        "--check",
+        help=(
+            "When polling is enabled this option can be used to check the"
+            " outcome of a given task. If the task fails the client will"
+            " fail too."
+        ),
+        action="store_true",
+    )
     parser_orchestrate.add_argument(
         "orchestrate_files",
         help="YAML files to use for orchestration.",
@@ -154,19 +221,8 @@ def _args():
     parser_exec = subparsers.add_parser("exec", help="Execution mode help")
     parser_exec.add_argument(
         "--verb",
-        help="Module Invocation for exec. Choices: %(choices)s",
+        help="Module Invocation for exec.",
         metavar="STRING",
-        choices=[
-            "RUN",
-            "COPY",
-            "ADD",
-            "ARG",
-            "ENV",
-            "WORKDIR",
-            "CACHEFILE",
-            "CACHEEVICT",
-            "QUERY",
-        ],
         required=True,
     )
     parser_exec.add_argument(
@@ -186,6 +242,25 @@ def _args():
         help="Block on client return for the completion of executed jobs.",
         action="store_true",
     )
+    parser_exec.add_argument(
+        "--stream",
+        help="Stream the STDOUT|STDERR for tasks.",
+        action="store_true",
+    )
+    parser_exec.add_argument(
+        "--check",
+        help=(
+            "When polling is enabled this option can be used to check the"
+            " outcome of a given task. If the task fails the client will"
+            " fail too."
+        ),
+        action="store_true",
+    )
+    parser_exec.add_argument(
+        "--force-async",
+        help=("Instruct the execution engine to run asynchronously."),
+        action="store_true",
+    )
     parser_server = subparsers.add_parser("server", help="Server mode help")
     parser_server.add_argument(
         "--bind-address",
@@ -193,19 +268,6 @@ def _args():
         metavar="STRING",
         default=os.getenv("DIRECTORD_BIND_ADDRESS", "*"),
     )
-    parser_server.add_argument(
-        "--etcd-server",
-        help="Domain or IP address of the ETCD server. Default: %(default)s",
-        metavar="STRING",
-        default=os.getenv("DIRECTORD_ETCD_SERVER", "localhost"),
-    )
-    parser_server.add_argument(
-        "--etcd-port",
-        help="ETCD server bind port. Default: %(default)s",
-        metavar="INT",
-        default=int(os.getenv("DIRECTORD_ETCD_PORT", 2379)),
-        type=int,
-    )
     parser_server.add_argument(
         "--run-ui",
         help="Enable the Directord UI. Default: %(default)s",
@@ -231,6 +293,12 @@ def _args():
     parser_manage = subparsers.add_parser(
         "manage", help="Server management mode help"
     )
+    filter_group = parser_manage.add_mutually_exclusive_group(required=False)
+    filter_group.add_argument(
+        "--filter",
+        choices=["success", "failed", "processing"],
+        help="List filtered jobs.",
+    )
     manage_group = parser_manage.add_mutually_exclusive_group(required=True)
     manage_group.add_argument(
         "--list-jobs", action="store_true", help="List all known jobs."
@@ -268,6 +336,21 @@ def _args():
         action="store_true",
         help="Generate encryption keys for Curve authentication.",
     )
+    manage_group.add_argument(
+        "--dump-cache",
+        action="store_true",
+        help="Dump the local cache to stdout.",
+    )
+    manage_group.add_argument(
+        "--analyze-parent",
+        help="Analyze a given parent ID.",
+        metavar="STRING",
+    )
+    manage_group.add_argument(
+        "--analyze-job",
+        help="Analyze a given job ID.",
+        metavar="STRING",
+    )
     parser_bootstrap = subparsers.add_parser(
         "bootstrap",
         help=(
@@ -281,6 +364,7 @@ def _args():
         help="File path for SSH catalog.",
         metavar="STRING",
         action="append",
+        required=True,
         type=argparse.FileType(mode="r"),
     )
     parser_bootstrap.add_argument(
@@ -299,38 +383,46 @@ def _args():
         default=int(os.getenv("DIRECTORD_BOOTSTRAP_THREADS", 10)),
         type=int,
     )
-    args = parser.parse_args()
+    if exec_args:
+        args = parser.parse_args(args=exec_args)
+    else:
+        args = parser.parse_args()
     # Check for configuration file and load it if found.
     if args.config_file:
         config_data = yaml.safe_load(args.config_file)
-        for key, value in config_data.items():
-            if isinstance(value, list) and key in args.__dict__:
-                args.__dict__[key].extend(value)
-            else:
-                args.__dict__[key] = value
+        if config_data:
+            for key, value in config_data.items():
+                if isinstance(value, list) and key in args.__dict__:
+                    args.__dict__[key].extend(value)
+                else:
+                    args.__dict__[key] = value
 
     return args, parser
 
 
-class SystemdInstall(object):
+class SystemdInstall:
     """Simple system service unit creation class."""
 
-    def __init__(self):
+    def __init__(self, group="root", force=False):
         """Class to create systemd service units.
 
         This class is used with the directord-server-systemd and
         directord-client-systemd entrypoints.
+
+        :param group: Name of the group used within the systemd service units.
+        :type group: String
+        :param force: Force install systemd service units.
+        :type force: Boolean
         """
 
         self.config_path = "/etc/directord"
+        self.socket_group = group
+        self.force = force
 
     def path_setup(self):
         """Create the configuration path and basic configuration file."""
 
-        if not os.path.isdir(self.config_path):
-            os.makedirs(self.config_path)
-            print("[+] Created directord configuration path")
-
+        os.makedirs(self.config_path, exist_ok=True)
         if not os.path.exists("/etc/directord/config.yaml"):
             with open("/etc/directord/config.yaml", "w") as f:
                 f.write("---\ndebug: false\n")
@@ -348,20 +440,28 @@ class SystemdInstall(object):
         self.path_setup()
         base = os.path.dirname(directord.__file__)
         service_file_path = "/etc/systemd/system/{}".format(service_file)
-        if os.path.exists(service_file_path):
+        if os.path.exists(service_file_path) and not self.force:
             print(
                 "[-] Service file was not created because it already exists."
             )
             return
-        with open(os.path.join(base, "static", service_file)) as f:
-            with open(service_file_path, "w") as service_f:
-                for line in f.readlines():
-                    service_f.write(
-                        line.replace(
-                            "/usr/bin/directord",
-                            os.path.join(path, "directord"),
-                        )
-                    )
+
+        blueprintLoader = jinja2.FileSystemLoader(
+            searchpath=os.path.join(base, "templates")
+        )
+        blueprintEnv = jinja2.Environment(
+            loader=blueprintLoader,
+            keep_trailing_newline=True,
+            undefined=StrictUndefined,
+        )
+        blueprint = blueprintEnv.get_template("{}.j2".format(service_file))
+        blueprint_args = {
+            "directord_binary": os.path.join(path, "directord"),
+            "directord_group": self.socket_group,
+        }
+        outputText = blueprint.render(**blueprint_args)
+        with open(service_file_path, "w") as f:
+            f.write(outputText)
 
         print("[+] Installed {} service unit file".format(service_file))
         print(
@@ -379,17 +479,41 @@ class SystemdInstall(object):
         self.writer(service_file="directord-client.service")
 
 
+def _systemd_loader():
+    parser = argparse.ArgumentParser(
+        description="Systemd install args.",
+        prog="Directord-service",
+        formatter_class=lambda prog: argparse.HelpFormatter(
+            prog, max_help_position=32, width=128
+        ),
+    )
+    parser.add_argument(
+        "--group",
+        help="Server group. Default: %(default)s.",
+        metavar="STRING",
+        default="root",
+        type=str,
+    )
+    parser.add_argument(
+        "--force",
+        help="Force install systemd service unit file.",
+        action="store_true",
+    )
+    args = parser.parse_args()
+    return SystemdInstall(group=args.group, force=args.force)
+
+
 def _systemd_server():
     """Execute the systemd server unit file creation process."""
 
-    _systemd = SystemdInstall()
+    _systemd = _systemd_loader()
     _systemd.server()
 
 
 def _systemd_client():
     """Execute the systemd client unit file creation process."""
 
-    _systemd = SystemdInstall()
+    _systemd = _systemd_loader()
     _systemd.client()
 
 
@@ -408,9 +532,9 @@ def main():
     _mixin = mixin.Mixin(args=args)
 
     if args.mode == "server":
-        _mixin.start_server()
+        server.Server(args=args).worker_run()
     elif args.mode == "client":
-        _mixin.start_client()
+        client.Client(args=args).worker_run()
     elif args.mode in ["exec", "orchestrate"]:
         if args.mode == "exec":
             return_data = _mixin.run_exec()
@@ -419,11 +543,40 @@ def main():
 
         job_items = [i.decode() for i in return_data if i]
 
-        if args.poll:
+        if args.poll or args.stream:
+            failed = list()
             manage = user.Manage(args=args)
             for item in job_items:
-                _, status = manage.poll_job(job_id=item)
+                state, status, stdout, stderr, info = manage.poll_job(
+                    job_id=item
+                )
+                if args.check and state is False:
+                    failed.append(item)
+
+                if args.stream:
+                    for node in sorted(
+                        set(
+                            i for v in [stdout, stderr, info] for i in v.keys()
+                        )
+                    ):
+                        for k, n, v in [
+                            (node, name, d[node])
+                            for name, d in [
+                                ("STDOUT", stdout),
+                                ("STDERR", stderr),
+                                ("INFO", info),
+                            ]
+                            if node in d
+                        ]:
+                            print("{} -- {} {}".format(k, n, v))
+
                 print(status)
+
+            if any(failed):
+                print("FAILED JOBS")
+                for item in failed:
+                    print(item)
+                raise SystemExit(1)
         else:
             for item in job_items:
                 print(item)
@@ -436,11 +589,17 @@ def main():
                 "Keys generated. Synchronize the server and client public"
                 " keys to client nodes to enable Curve encryption."
             )
-
-        if not data:
             return
 
-        data = json.loads(data)
+        try:
+            data = json.loads(data)
+        except Exception as e:
+            print("No valid data found: {}".format(str(e)))
+            return
+        else:
+            if not data:
+                raise SystemExit("No data found")
+
         if args.export_jobs or args.export_nodes:
             export_file = utils.dump_yaml(
                 file_path=(args.export_jobs or args.export_nodes),
@@ -450,12 +609,19 @@ def main():
             return
 
         computed_values = dict()
-        if data and isinstance(data, list):
+        headings = ["KEY", "VALUE"]
+        tabulated_data = None
+        if data and isinstance(data, dict):
+            tabulated_data = _mixin.return_tabulated_info(data=data)
+        elif data and isinstance(data, list):
             if args.job_info:
-                headings = ["KEY", "VALUE"]
-
                 item = dict(data).get(args.job_info)
                 if not item:
+                    print(
+                        "Job information for ID:{} was not found".format(
+                            args.job_info
+                        )
+                    )
                     return
 
                 tabulated_data = _mixin.return_tabulated_info(data=item)
@@ -463,22 +629,30 @@ def main():
                 if args.list_jobs:
                     restrict_headings = [
                         "PARENT_JOB_ID",
+                        "VERB",
                         "EXECUTION_TIME",
+                        "PROCESSING",
                         "SUCCESS",
                         "FAILED",
                     ]
                 else:
                     restrict_headings = [
                         "EXPIRY",
+                        "VERSION",
+                        "HOST_UPTIME",
+                        "AGENT_UPTIME",
                     ]
-                _tabulated_data = _mixin.return_tabulated_data(
+                (
+                    tabulated_data,
+                    headings,
+                    computed_values,
+                ) = _mixin.return_tabulated_data(
                     data=data, restrict_headings=restrict_headings
                 )
-                tabulated_data, headings, computed_values = _tabulated_data
-            print(
-                tabulate.tabulate(
-                    [i for i in tabulated_data if i], headers=headings
-                )
+
+        if tabulated_data:
+            utils.print_tabulated_data(
+                data=[i for i in tabulated_data if i], headers=headings
             )
             print("\nTotal Items: {}".format(len(tabulated_data)))
             for k, v in computed_values.items():
@@ -486,8 +660,11 @@ def main():
                     print("Total {}: {:.2f}".format(k, v))
                 else:
                     print("Total {}: {}".format(k, v))
+            else:
+                return
     elif args.mode == "bootstrap":
-        _mixin.bootstrap_cluster()
+        _bootstrap = bootstrap.Bootstrap(args=args)
+        _bootstrap.bootstrap_cluster()
     else:
         parser.print_help(sys.stderr)
         raise SystemExit("Mode is set to an unsupported value.")
